
import peasy.*;

PeasyCam cam;

boolean[][] cellsNow; //2 dimensional array of bools
boolean[][] cellsFuture; //2 dimensional array of bools

boolean[][][] history; //this is a 3dimensional array used to store the history of the CA
int historyLength=60; //the maximum amount of steps to store
int currentTime=0;

int columns, rows;
int scale=5;

//Wolfram rule 746
boolean[] trueRules={
  true,
  true,
  true,
  true,
  true,
  false,
  false,
  false,
  false,
  false,
};
boolean[] falseRules={
  false,
  false,
  false,
  true,
  false,
  false,
  false,
  false,
  false,
  false,
};

void setup(){

size(500,500, P3D);
pixelDensity(displayDensity());
cam = new PeasyCam(this, 400);
background(255); //draw background white
columns=80; 
rows=80;

cellsNow=new boolean[columns][rows]; 
cellsFuture=new boolean[columns][rows];

history=new boolean[columns][rows][historyLength];

//this sets a row of 3 contiguous cells at the centre as true
cellsNow[columns/2][rows/2]=true;
cellsNow[columns/2][rows/2-1]=true;
cellsNow[columns/2][rows/2+1]=true;
}

void draw(){

  background(255); //clean the background
  
  //set some lighting...
  ambientLight(30,30,30);
  lightSpecular(255,255,255);
  directionalLight(255,255,255,-30,60,-50);
 
  //to center the CA and make it look nice at the start
  rotateY(-2*PI/3);
  rotateX(-PI/3);
  translate(-scale*columns/2,-scale*rows/2,-scale*historyLength/2); 
  
  drawHistory(); //draw
  tick(); //calculate future values
  recordHistory();
  update(); 
}

void drawHistory(){
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      for(int k=0;k<currentTime;k++){
        if(history[i][j][k]==true){ //draw only if true
          pushMatrix();
          translate(i*scale,j*scale,k*scale);
          //sphere(scale/1.5);
          box(scale,scale,scale);
          popMatrix(); //needs to be called eventually after pushMatrix().
        }
      }
    }
  }
}


void tick(){
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      int numNeigs= countNeighbours(i,j);
       cellsFuture[i][j]=applyRules(cellsNow[i][j],numNeigs);
    }
  }
}

int countNeighbours(int col,int row){

  int totalNeighbours=0;
  for(int i=col-1;i<=col+1;i++){
    //wrap the column.
    int neigColumn=wrap(i,columns);
    for(int j=row-1;j<=row+1;j++){
      //wrap the row.
      int neigRow=wrap(j,rows);
      // if this neighbour is "on" (true) we add it
      if(cellsNow[neigColumn][neigRow]==true){
        totalNeighbours++;
      }
    } //end of row loop
  } //end of column loop
  return totalNeighbours;

}

//this is the wrap function.
int wrap(int index,int maxBound){
  int result;
  if(index<0){
     result=maxBound+index;
  }
  else if(index>=maxBound){
    result=index-maxBound;
  }
  else{
    result=index;
  }
  return result;
}

boolean applyRules(boolean me, int numNeigs){
  if(me==true){
    return trueRules[numNeigs];
  }else {
    return falseRules[numNeigs];
  } 
}

void update(){
  //copying all values.
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      cellsNow[i][j]=cellsFuture[i][j];
    }
  }
}

void recordHistory(){
  //to prevent storing beyond the capacity of history
  if(currentTime>=historyLength){ 
    return;
  }
  //if it reaches here, currentTime is smaller than historyLength.

  //store the current states in history, at currentTime
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      history[i][j][currentTime]=cellsNow[i][j];    
    }
  }
  currentTime++; //increase current time
}