import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import peasy.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class Exercise19CA3D extends PApplet {




PeasyCam cam;

boolean[][] cellsNow; //2 dimensional array of bools
boolean[][] cellsFuture; //2 dimensional array of bools

boolean[][][] history; //this is a 3dimensional array used to store the history of the CA
int historyLength=60; //the maximum amount of steps to store
int currentTime=0;

// things clearer
int columns, rows;
int scale=5; //this means that the cells will be 5x5 units now...

//these are the rules for a von Neumann neighborhood
//(there are a maximum of 4 neighbours on)

boolean[] trueRules=new boolean[10];
boolean[] falseRules=new boolean[10];


//this is used just to slow down the drawing
int drawCounts=0;

public void setup(){



cam = new PeasyCam(this, 100);

background(255); //draw background white
//stroke for showing the cells
noStroke();
fill(200);

columns=80; 
rows=80;

cellsNow=new boolean[columns][rows]; 
cellsFuture=new boolean[columns][rows];

history=new boolean[columns][rows][historyLength];

//this sets a row of 5 contiguous cells at the centre as true
cellsNow[columns/2][rows/2]=true;
// cellsNow[columns/2][rows/2-4]=true;
// cellsNow[columns/2][rows/2-3]=true;
// cellsNow[columns/2][rows/2-2]=true;
cellsNow[columns/2][rows/2-1]=true;
cellsNow[columns/2][rows/2+1]=true;
// cellsNow[columns/2][rows/2+2]=true;
// cellsNow[columns/2][rows/2+3]=true;
// cellsNow[columns/2][rows/2+4]=true;

makeRulesFromWolfram(746);

}

public void draw(){

  background(255); //clean the background
  
  //set some lighting...
  ambientLight(30,30,30);
  lightSpecular(255,255,255);
  directionalLight(255,255,255,-5.0f,13,28);
 
  //to center the drawing
  translate(-scale*columns/2,-scale*rows/2,-scale*historyLength/2); 

  drawHistory(); //draw
  //to slow down the drawing this keeps tract of how many times Processing has called draw()
  //and only draws every 10th time (see the % modulo operator in the reference)
  if(drawCounts%10==0){ 
    tick(); //calculate future values
    recordHistory();
  }
  drawCounts++;//increase the drawCounts...

  update(); 
}

public void drawHistory(){
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      for(int k=0;k<currentTime;k++){
        if(history[i][j][k]==true){ //draw only if true
          pushMatrix();
          translate(i*scale,j*scale,k*scale);
          //sphere(scale/1.5);
          box(scale,scale,scale);
          popMatrix(); //needs to be called eventually after pushMatrix().
        }
      }
    }
  }
}


public void tick(){
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      int numNeigs= countNeighbours(i,j);
       cellsFuture[i][j]=applyRules(cellsNow[i][j],numNeigs);
    }
  }
}

public int countNeighbours(int col,int row){

  int totalNeighbours=0;
  for(int i=col-1;i<=col+1;i++){
    //wrap the column.
    int neigColumn=wrap(i,columns);
    for(int j=row-1;j<=row+1;j++){
      //wrap the row.
      int neigRow=wrap(j,rows);
      // if this neighbour is "on" (true) we add it
      if(cellsNow[neigColumn][neigRow]==true){
        totalNeighbours++;
      }
    } //end of row loop
  } //end of column loop
  return totalNeighbours;

}

//this is the wrap function.
public int wrap(int index,int maxBound){
  int result;
  if(index<0){
     result=maxBound+index;
  }
  else if(index>=maxBound){
    result=index-maxBound;
  }
  else{
    result=index;
  }
  return result;
}


public boolean applyRules(boolean me, int numNeigs){
  if(me==true){
    return trueRules[numNeigs];
  }else {
    return falseRules[numNeigs];
  } 
}


public void update(){
  //copying all values.
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      cellsNow[i][j]=cellsFuture[i][j];
    }
  }
}

public void recordHistory(){
  //to prevent storing beyond the capacity of history
  if(currentTime>=historyLength){ 
    return;
  }
  //if it reaches here, currentTime < historyLength.

//store the current states in history, at currentTime
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      history[i][j][currentTime]=cellsNow[i][j];    
    }
  }
  currentTime++; //increase current time
}

public void makeRulesFromWolfram(int rule){ //generate rules from Wolfram encoding
  for(int i=0;i<10;i++){
    falseRules[i] = (rule & 1<<(2*i))!=0;
    trueRules[i]  = (rule & 1<<(2*i+1))!=0;
  }
  println("//Wolfram rule "+rule);
  println("boolean[] trueRules={");
  for(int i=0;i<10;i++){
    println("\t"+trueRules[i] +",");
  }
  println("};");
  println("boolean[] falseRules={");
  for(int i=0;i<10;i++){
    println("\t"+falseRules[i] +",");
  }
  println("};");
}
  public void settings() { 
size(500,500, P3D); 
pixelDensity(displayDensity()); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--present", "--window-color=#666666", "--hide-stop", "Exercise19CA3D" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
